//! The query builder API
//!
//! [`Metric`] and [`Scalar`] are the basic types used for building queries.
//! For building complex queries involving operators, see module [`ops`].

use std::fmt::Display;

use crate::seal::Sealed;

use self::ops::Operable;

pub mod fns;
pub mod ops;

// Add default implementation of IntoQuery where the type simply returns itself
// Used for types that already implement AsRef<str>
macro_rules! impl_into_query {
    ($($t:ty),+) => {
        $(
            impl IntoQuery for $t {
                type Target = Self;
                fn into_query(self) -> Self::Target {
                    self
                }
            }
        )+
    };
}

/// Marker trait to signify that a type is queriable
///
/// Has a blanket implementation for all types that implement `AsRef<str>`
pub trait Query: AsRef<str> {}

impl<T> Query for T where T: AsRef<str> {}

/// Used by non-string types provided by the crate to convert into a type implements [``Query``]
///
/// For allowing raw queries, `IntoQuery` has been implemented for `String`, `&String`, and `&str`.
pub trait IntoQuery {
    type Target: Query;
    fn into_query(self) -> Self::Target;
}

impl_into_query!(String, &String, &str);

/// Metric selector/builder. Used to building queries with the help of the type system.
///
/// Note: besides making use of the type system, this type doesn't do any kind of validation on the resulting queries.
#[derive(Default, Debug, Clone)]
pub struct Metric<'a> {
    name: &'a str,
    labels: Vec<Label<'a>>,
    at: Option<i32>,
    lookback: Option<Duration>,
    offset: Option<Offset>,
}

impl<'a> Metric<'a> {
    /// Create a new metric with a metric name (for example `total_http_requests`)
    ///
    /// A metric can be constructed with an empty string as well, however, it's important to
    /// have at least one label matcher present for the query to be valid.
    pub fn new(name: &'a str) -> Self {
        Self {
            name,
            ..Default::default()
        }
    }

    /// Add a label matcher with an equality operator i.e. name="value"
    ///
    /// For addding multiple labels at once, see [`Self::labels`]
    pub fn label_eq(mut self, name: &'a str, value: &'a str) -> Self {
        self.labels.push(Label::Eq(name, value));
        self
    }

    /// Add a label matcher with an inequality operator i.e. name!="value"
    ///
    /// For addding multiple labels at once, see [`Self::labels`]
    pub fn label_ne(mut self, name: &'a str, value: &'a str) -> Self {
        self.labels.push(Label::Ne(name, value));
        self
    }

    /// Add a regular expression label matcher with an equality operator i.e. name=~".+"
    ///
    /// For addding multiple labels at once, see [`Self::labels`]
    pub fn label_rx_eq(mut self, name: &'a str, value: &'a str) -> Self {
        self.labels.push(Label::RxEq(name, value));
        self
    }

    /// Add a regular expression label matcher with an equality operator i.e. name!~".+"
    ///
    /// For addding multiple labels at once, see [`Self::labels`]
    pub fn label_rx_ne(mut self, name: &'a str, value: &'a str) -> Self {
        self.labels.push(Label::RxNe(name, value));
        self
    }

    /// Add multiple label matchers at once
    ///
    /// # Example
    /// ```rust
    /// use mquery::query::{Metric, Label};
    ///
    /// let _ = Metric::new("metric_name")
    ///     .labels([
    ///         Label::Eq("name", "value"),
    ///         Label::Ne("name", "value"),
    ///         Label::RxEq("name", "value"),
    ///     ]);
    /// ```
    pub fn labels<I: IntoIterator<Item = Label<'a>>>(mut self, labels: I) -> Self {
        self.labels.extend(labels);
        self
    }

    /// Set query evaluation time to the specified Unix timestamp using the `@` modifier.
    pub fn at(mut self, timestamp: i32) -> Self {
        // TODO: Convert between chrono DataTime
        self.at.replace(timestamp);
        self
    }

    /// Set lookback duration / range for the query with the appropriate unit of time
    ///
    /// More info on range selectors can be found at [Range Vector Selectors](https://prometheus.io/docs/prometheus/latest/querying/basics/#range-vector-selectors)
    pub fn lookback(mut self, duration: Duration) -> Self {
        self.lookback.replace(duration);
        self
    }

    /// Set the offset time fromt the query evaluation time using the offset modifier with the appropriate unit
    ///
    /// Note: Offsets can contain [negative values](https://prometheus.io/docs/prometheus/latest/querying/basics/#offset-modifier)
    pub fn offset(mut self, offset: Offset) -> Self {
        self.offset.replace(offset);
        self
    }
}

impl Display for Metric<'_> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str(self.name)?;
        if !self.labels.is_empty() {
            f.write_str("{")?;
            for (i, l) in self.labels.iter().enumerate() {
                write!(f, "{l}")?;
                if i != (self.labels.len() - 1) {
                    f.write_str(",")?;
                }
            }
            f.write_str("}")?;
        }
        if let Some(lookback) = self.lookback {
            write!(f, "{lookback}")?;
        }
        if let Some(at) = self.at {
            write!(f, " @ {}", at)?;
        }
        if let Some(off) = self.offset {
            write!(f, " {off}")?;
        }

        Ok(())
    }
}

impl IntoQuery for Metric<'_> {
    type Target = String;
    fn into_query(self) -> Self::Target {
        self.to_string()
    }
}

impl Sealed for Metric<'_> {}
impl ops::Operable for Metric<'_> {}

/// Label matcher type with (Name, Value) pairs
#[derive(Debug, Clone)]
pub enum Label<'a> {
    Eq(&'a str, &'a str),
    Ne(&'a str, &'a str),
    RxEq(&'a str, &'a str),
    RxNe(&'a str, &'a str),
}

impl Display for Label<'_> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Label::Eq(n, v) => write!(f, "{n}=\"{v}\"")?,
            Label::Ne(n, v) => write!(f, "{n}!=\"{v}\"")?,
            Label::RxEq(n, v) => write!(f, "{n}=~\"{v}\"")?,
            Label::RxNe(n, v) => write!(f, "{n}!~\"{v}\"")?,
        }
        Ok(())
    }
}

/// Units of time to be used with [``Duration``] or [``Offset``]
#[derive(Debug, Clone, Copy)]
pub enum Unit {
    Ms,
    Sec,
    Min,
    Hr,
    Day,
    Week,
    Year,
}

impl Display for Unit {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Unit::Ms => write!(f, "ms"),
            Unit::Sec => write!(f, "s"),
            Unit::Min => write!(f, "m"),
            Unit::Hr => write!(f, "h"),
            Unit::Day => write!(f, "d"),
            Unit::Week => write!(f, "w"),
            Unit::Year => write!(f, "y"),
        }
    }
}

/// Specify lookback / range duration for the queries
#[derive(Debug, Clone, Copy)]
pub struct Duration(pub u64, pub Unit);

impl Display for Duration {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "[{}{}]", self.0, self.1)
    }
}

impl From<(u64, Unit)> for Duration {
    fn from((time, unit): (u64, Unit)) -> Self {
        Duration(time, unit)
    }
}

/// Specify time offset for the queries
#[derive(Debug, Clone, Copy)]
pub struct Offset(i64, Unit);

impl Display for Offset {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "offset {}{}", self.0, self.1)
    }
}

/// A scalar value that represents and also holds double precision floating point numbers
///
/// This is one of the two base types for building typed queries. Another is [``Metric``].
/// Scalar implements [``Operable``], which means that it supports the operations enabled by the
/// [``ops``] module.
#[derive(Debug, Clone, Copy, Default)]
pub struct Scalar(pub f64);

impl Display for Scalar {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        self.0.fmt(f)
    }
}

impl Sealed for Scalar {}
impl ops::Operable for Scalar {}

/// Raw Expression type that allows raw strings to be used in the builder API where necessary
///
/// This also allows combining complex queries, especially when a function is generic over a single type
/// `T` that implements `Operable`, you can convert the existing type into this to make it work.
/// This can be done easily with the [`QueryExt`] trait which has a blanket implementation for all types
/// which implement [`Operable`] (all query builder API types).
///
/// Note: This is generally not recommended unless it is necessary to use. This is because the query builder
/// API is lazy. It doesn't build the query until `to_string` or `into_query` methods are called. This avoids
/// many intermediate string allocations. The [`RawExpr`] type, however, is built with a `String`, which means
/// that when you turn `T: Operable` into [`RawExpr`], the `to_string` method is called early.
#[derive(Debug, Clone, PartialEq, Eq)]
pub struct RawExpr(String);

impl RawExpr {
    /// Construct a raw expression type from a string
    ///
    /// See type level docs for additional details
    pub fn new<T: ToOwned<Owned = String>>(query: T) -> Self {
        RawExpr(query.to_owned())
    }
    /// Construct a raw expression type from a type that that implements `Operable`
    ///
    /// See type level docs for additional details
    pub fn from_expr<E: Operable>(expr: E) -> Self {
        RawExpr(expr.to_string())
    }
    /// Get the inner string
    pub fn into_inner(self) -> String {
        self.0
    }
}

impl AsRef<str> for RawExpr {
    fn as_ref(&self) -> &str {
        self.0.as_str()
    }
}

impl From<&str> for RawExpr {
    fn from(value: &str) -> Self {
        RawExpr(value.to_owned())
    }
}

impl From<String> for RawExpr {
    fn from(value: String) -> Self {
        RawExpr(value)
    }
}

impl IntoQuery for RawExpr {
    type Target = String;
    fn into_query(self) -> Self::Target {
        self.0
    }
}

impl Operable for RawExpr {}
impl Sealed for RawExpr {}

impl Display for RawExpr {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        self.0.fmt(f)
    }
}

/// Additional/extended query builder API functionality
///
/// This trait has a blanket implementation for all types that implement [`Operable`]
pub trait QueryExt {
    /// Convert a query builder API type into a [`RawExpr`] type
    ///
    /// This is useful when a function takes a sequence (array/vector/iterator) of a generic type `T: Operable`
    /// but you want to combine different types (produced from query fns or other complex types).
    ///
    /// While the `String` type could have been allowed to be part of the builder API directly,
    /// it is by design that a user must explicitly turn a string into a [`RawExpr`] saying that
    /// they're opting in to use a raw string in the query builder API.
    fn to_expr(&self) -> RawExpr;
    /// Call a function or closure on `Self` to transform it into another type
    ///
    /// This is similar to the `Iterator's` `map` method, except this is eagerly evaluated and operates on a single item.
    /// This is useful for building complex queries that have many nested functions.
    ///
    /// For example:
    /// ```
    /// # use mquery::query::{Metric, QueryExt, fns::aggregate::sum};
    /// // the usual way
    /// let nest = sum(Metric::new("http_requests"));
    /// // with `then`
    /// let chain = Metric::new("http_requests").then(sum);
    /// assert_eq!(nest.to_string(), chain.to_string());
    /// ```
    ///
    /// > Note: the reason this function is called `then` instead of `pipe` or `map` is because `then` semantically provides a clearer
    /// > idea of what's happening when writing/reading the complex queries.
    fn then<R>(self, func: impl FnOnce(Self) -> R) -> R
    where
        Self: Sized;
}

impl<T: Operable> QueryExt for T {
    fn to_expr(&self) -> RawExpr {
        RawExpr(self.to_string())
    }

    fn then<R>(self, func: impl FnOnce(Self) -> R) -> R
    where
        Self: Sized,
    {
        func(self)
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn metric_selector() {
        let metric = Metric::new("metric_name")
            .labels([
                Label::Eq("name", "value"),
                Label::Ne("name", "value"),
                Label::RxEq("name", "value"),
            ])
            .label_rx_ne("name", "value")
            .at(123456)
            .lookback(Duration(1, Unit::Year))
            .offset(Offset(-1, Unit::Year))
            .to_string();

        assert_eq!(
            r#"metric_name{name="value",name!="value",name=~"value",name!~"value"}[1y] @ 123456 offset -1y"#,
            metric
        )
    }
}
