use std::fmt::Display;

use crate::{query::IntoQuery, seal::Sealed};

use super::{
    modifiers::{GroupType, MatchType, Mod},
    Operable,
};

/// Trait that enables arithmetic query operations on all types that implement [``Operable``]
/// (i.e. [``super::Scalar``] and [``super::Metric``] )
pub trait Arithmetic<'a>: Operable
where
    Self: Sized,
{
    /// Add `Self` with another operable type to produce a query that looks like `(self + right)`
    fn add<R: Operable>(self, right: R) -> ArithmeticOperation<'a, Self, R> {
        ArithmeticOperation::new(ArthOp::Add, self, right)
    }
    /// Subtract `Self` with another operable type to produce a query that looks like `(self - right)`
    fn sub<R: Operable>(self, right: R) -> ArithmeticOperation<'a, Self, R> {
        ArithmeticOperation::new(ArthOp::Sub, self, right)
    }
    /// Multiply `Self` with another operable type to produce a query that looks like `(self * right)`
    fn mul<R: Operable>(self, right: R) -> ArithmeticOperation<'a, Self, R> {
        ArithmeticOperation::new(ArthOp::Mul, self, right)
    }
    /// Divide `Self` with another operable type to produce a query that looks like `(self / right)`
    fn div<R: Operable>(self, right: R) -> ArithmeticOperation<'a, Self, R> {
        ArithmeticOperation::new(ArthOp::Div, self, right)
    }
    /// Apply modulo operator between `Self` and another operable type to produce a query that looks like `(self % right)`
    fn modulo<R: Operable>(self, right: R) -> ArithmeticOperation<'a, Self, R> {
        ArithmeticOperation::new(ArthOp::Mod, self, right)
    }

    /// Apply the exponent/power operator between `Self` and another operable type to produce a query that looks like `(self ^ right)`
    fn pow<R: Operable>(self, right: R) -> ArithmeticOperation<'a, Self, R> {
        ArithmeticOperation::new(ArthOp::Pow, self, right)
    }
}

impl<T> Arithmetic<'_> for T where T: Operable {}

/// Represents an arithmetic operation between type `A` and type `B`
/// where `A` and `B` both implement [``Operable``]
///
/// This struct is not meant to be constructed directly by the user of the library.
/// Queries with arithmetic operations should be constructed through the [``Arithmetic``] trait.
#[derive(Debug, Clone)]
pub struct ArithmeticOperation<'a, A, B>
where
    A: Operable,
    B: Operable,
{
    op_type: ArthOp,
    left: A,
    right: B,
    matching: Mod<'a, MatchType>,
    group_mod: Mod<'a, GroupType>,
}

impl<'a, A, B> ArithmeticOperation<'a, A, B>
where
    A: Operable,
    B: Operable,
{
    fn new(op_type: ArthOp, left: A, right: B) -> Self {
        Self {
            op_type,
            left,
            right,
            matching: Default::default(),
            group_mod: Default::default(),
        }
    }
    /// Apply `group_left` group modifier to the arithmetic operation without specifying any label names
    pub fn group_left(mut self) -> Self {
        self.group_mod = Mod::from((GroupType::Left, []));
        self
    }
    /// Apply `group_left` group modifier to the arithmetic operation with label names
    pub fn group_left_labels(mut self, labels: impl IntoIterator<Item = &'a str>) -> Self {
        self.group_mod = Mod::from((GroupType::Left, labels));
        self
    }
    /// Apply `group_right` group modifier to the arithmetic operation without specifying any label names
    pub fn group_right(mut self) -> Self {
        self.group_mod = Mod::from((GroupType::Right, []));
        self
    }
    /// Apply `group_right` group modifier to the arithmetic operation with label names
    pub fn group_right_labels(mut self, labels: impl IntoIterator<Item = &'a str>) -> Self {
        self.group_mod = Mod::from((GroupType::Right, labels));
        self
    }
    /// Match only given labels between the two vectors for evauluating the arithmetic operation
    ///
    /// Note: Only to be used with vector expressions on both left and right
    /// i.e. `<vector expr> <bin-op> on(<label list>) <vector expr>`
    pub fn match_on(mut self, labels: impl IntoIterator<Item = &'a str>) -> Self {
        self.matching = Mod::from((MatchType::On, labels));
        self
    }

    /// Ignore given labels between the two vectors for evauluating the arithmetic operation
    ///
    /// Note: Only to be used with vector expressions on both left and right
    /// i.e. `<vector expr> <bin-op> ignoring(<label list>) <vector expr>`
    pub fn ignoring(mut self, labels: impl IntoIterator<Item = &'a str>) -> Self {
        self.matching = Mod::from((MatchType::Ignoring, labels));
        self
    }
}

impl<A, B> Operable for ArithmeticOperation<'_, A, B>
where
    A: Operable,
    B: Operable,
{
}

impl<A, B> Sealed for ArithmeticOperation<'_, A, B>
where
    A: Operable,
    B: Operable,
{
}

#[derive(Clone, Copy, Debug)]
enum ArthOp {
    Add = '+' as isize,
    Sub = '-' as isize,
    Mul = '*' as isize,
    Div = '/' as isize,
    Mod = '%' as isize,
    Pow = '^' as isize,
}

impl ArthOp {
    fn to_char(self) -> char {
        self as u8 as char
    }
}

impl<A, B> Display for ArithmeticOperation<'_, A, B>
where
    A: Operable,
    B: Operable,
{
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let op = self.op_type.to_char();
        write!(
            f,
            "({left} {op}{matching}{group} {right})",
            left = self.left,
            right = self.right,
            matching = self.matching,
            group = self.group_mod,
        )
    }
}

impl<A, B> IntoQuery for ArithmeticOperation<'_, A, B>
where
    A: Operable,
    B: Operable,
{
    type Target = String;
    fn into_query(self) -> Self::Target {
        self.to_string()
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use crate::query::{Metric, Scalar};

    #[test]
    fn arith_all_ops() {
        let query = Scalar(1.0);
        assert_eq!(query.add(query).to_string(), "(1 + 1)");
        assert_eq!(query.sub(query).to_string(), "(1 - 1)");
        assert_eq!(query.mul(query).to_string(), "(1 * 1)");
        assert_eq!(query.div(query).to_string(), "(1 / 1)");
        assert_eq!(query.modulo(query).to_string(), "(1 % 1)");
        assert_eq!(query.pow(query).to_string(), "(1 ^ 1)");
    }

    #[test]
    fn arith_scalar_scalar() {
        let query = Scalar(0.0).add(Scalar(1.0));
        assert_eq!(query.to_string(), "(0 + 1)");
    }

    #[test]
    fn arith_metric_scalar() {
        let query = Metric::new("test_metric").sub(Scalar(0.5));
        assert_eq!(query.to_string(), "(test_metric - 0.5)");
    }

    #[test]
    fn arith_nested() {
        let query = Metric::new("test_metric").sub(Scalar(0.5)).mul(Scalar(0.1));
        assert_eq!(query.to_string(), "((test_metric - 0.5) * 0.1)");
    }

    #[test]
    fn arith_match_on() {
        let metric = Metric::new("test_metric")
            .add(Metric::new("another_metric"))
            .match_on(["one", "two"])
            .to_string();
        assert_eq!(metric, "(test_metric + on(one,two) another_metric)");
    }

    #[test]
    fn arith_match_ignoring() {
        let metric = Metric::new("test_metric")
            .add(Metric::new("another_metric"))
            .ignoring(["one", "two"])
            .to_string();
        assert_eq!(metric, "(test_metric + ignoring(one,two) another_metric)");
    }

    #[test]
    fn arith_group_left() {
        let metric = Metric::new("test_metric")
            .add(Metric::new("another_metric"))
            .group_left()
            .to_string();
        assert_eq!(metric, "(test_metric + group_left another_metric)");
    }

    #[test]
    fn arith_group_right() {
        let metric = Metric::new("test_metric")
            .add(Metric::new("another_metric"))
            .group_right()
            .to_string();
        assert_eq!(metric, "(test_metric + group_right another_metric)");
    }

    #[test]
    fn arith_group_left_labels() {
        let metric = Metric::new("test_metric")
            .add(Metric::new("another_metric"))
            .group_left_labels(["label", "label2"])
            .to_string();
        assert_eq!(
            metric,
            "(test_metric + group_left(label,label2) another_metric)"
        );
    }

    #[test]
    fn arith_group_right_labels() {
        let metric = Metric::new("test_metric")
            .add(Metric::new("another_metric"))
            .group_right_labels(["label", "label2"])
            .to_string();
        assert_eq!(
            metric,
            "(test_metric + group_right(label,label2) another_metric)"
        );
    }
}
